/**
 * @file Ce fichier regroupe les différentes classes avec différents admins. Ces classes sont dédiées à être exportées directement pour être utilisées par le solver.
 * Le découpage par fichier est arbitraire mais permet de regrouper certaines classes proches.
 * @author hawkspar
 */

import LDAP from './basics';
import {SmartSearch, Tests} from './utilities';
import User from './users';
// Essentiels pour le fichier de config
import path from 'path';
import fs from 'fs';

// Point central ; tous les champs de la BDD sont 'cachés' dans config.json et pas visibles directement
var configPath = path.resolve('./','ldap_config.json');
var config = JSON.parse(fs.readFileSync(configPath, 'utf8'));

class Admin extends User {
    /**
     * @class Cette classe est la classe de l'administrateur d'un groupe qui lui permet de rajouter des membres, en supprimer, idem pour des admins,
     * ou éditer, voir supprimer le groupe.
     * @summary Ce constructeur appelle simplement le constructeur de sa classe mère.
     * @arg {Object} user - Utilisateur de la forme nécessaire à {@link LDAP.bind}.
    */
    constructor(user) { super(user); }
    
    //------------------------------------------------------------------------------------------------------------------------
    // Fonctions de relation TBT
    //------------------------------------------------------------------------------------------------------------------------
    
    /**
     * @summary Fonction qui permet de rajouter un membre déjà créé à un groupe.
     * @desc Cette fonction fait essentiellement appel à {@link LDAP.modifier} et {@link listerGroupes}. Elle n'autorise pas les doublons et opère dans les deux dns users
     * et groups.
     * @arg {string} uid - Identifiant du futur membre
     * @arg {string} gid - Identifiant du groupe
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     */
    static addGroupMember(uid, gid) {
        // Vérifie que l'utilisateur est pas déjà membre pour groupes
        return this.getMembers(gid).then(lm => {
            if (!lm.includes(uid)) {
                let vals = {};
                vals[config.groups.member] = uid;
                LDAP.change(config.key_id+gid+config.dn_groups, "add", vals).then(res => {
                    // Erreur si pb lors de la modification
                    if (!res) { throw Error; }
                });
            }
        // Vérifie que l'utilisateur est pas déjà membre pour users
        }).then(res => this.getGroups(uid)).then( lg => {
            if (!lg.includes(gid)) {
                let vals2 = {};
                vals2[config.users.groups] = gid;
                LDAP.change(config.key_id+uid+config.dn_users, "add", vals2).then(res => {
                    // Erreur si pb lors de la modification
                    if (!res) { throw Error; } else { return true; }
                });
            }
        });
    }

    /**
     * @summary Fonction qui permet de supprimer un membre existant d'un groupe.
     * @desc Cette fonction fait essentiellement appel à {@link LDAP.search}, {@link LDAP.change}, {@link Open.getGroups} et {@link Open.getMembers}.
     * @arg {string} uid - Identifiant de l'ex-membre
     * @arg {string} gid - Identifiant du groupe
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     */
    static delGroupMember(uid, gid) {
        // Vérifie que l'utilisateur est pas déjà viré pour groupes
        return this.getMembers(gid).then(lm => {
            if (lm.includes(uid)) {
                // Supprime tous les utilisateurs
                LDAP.change(config.key_id+gid+config.dn_groups, "del", config.group.member).then(res => {
                    // Les rajoute un par un, sauf pour le supprimé
                    lm.forEach(id => {
                        if (id!=uid) { this.addGroupMember(id, gid); }
                    });
                });
            }
        }).then(res => this.getGroups(uid)).then(lg => {
            // Vérifie que l'utilisateur est pas déjà viré pour users
            if (lg.includes(gid)) {
                // Supprime tous les groupes
                LDAP.change(config.key_id+uid+config.dn_users, "del", config.member.groups).then(res => {
                    // Les rajoute un par un, sauf pour le supprimé
                    lg.forEach(id => {
                        if (id!=gid) { this.addGroupMember(uid, id).then(res => { if (res) { return true; } else { throw Error; }}); }
                    });
                });
            }
        });
    }

    /**
     * @summary Fonction qui permet de promouvoir membre au stade d'administrateur d'un groupe.
     * @desc Cette fonction fait essentiellement appel à {@link Admin.addGroupMember} {@link LDAP.change} et {@link Open.getAdmins}. Elle n'autorise pas
     * les doublons et opère dans les deux dns users et groups.
     * @arg {string} uid - Identifiant du futur membre
     * @arg {string} gid - Identifiant du groupe
     * @return {boolean} `true` si la modification s'est bien déroulée, false sinon
     */
    static addGroupAdmin(uid, gid){
        // Ajoute le membre au groupe avant d'en faire un admin
        return this.addGroupMember(uid,gid).then(res => {
            this.getAdmins(gid).then(la => {
                if (!la.includes(uid)) {
                    // Finalement modification, uniquement dans groups
                    let vals = {};
                    vals[config.groups.admin] = uid;
                    LDAP.change(config.key_id+gid+config.dn_groups, "add", vals).then(res => {
                        // Gestion d'erreur
                        if (!res) { throw Error; } else { return true; }
                    });
                }
            });
        });
    }

    /**
     * @summary Fonction qui permet de rétrograder un membre du stade d'administrateur d'un groupe au stade d'utilisateur.
     * @desc Cette fonction fait essentiellement appel à {@link LDAP.change}.
     * @arg {string} uid - Identifiant du futur membre
     * @arg {string} gid - Identifiant du groupe
     * @return {boolean} `true` si la modification s'est bien déroulée, false sinon
     */
    delGroupAdmin(uid, gid) {
        // Peut paraître absurde mais permet de s'assurer que le membre est bien présent et que ses champs sont comme il faut
        return this.delGroupMember(uid, gid).then( res => { this.addGroupMember(uid,gid).then( res =>
            // Vérifie que l'utilisateur est bien admin (comme dans delGroupMember)
            this.getAdmins(gid).then(la => {
                if (la.includes(uid)) {
                    // Supprime tous les administrateurs
                    LDAP.change(config.key_id+gid+config.dn_groups, "del", config.group.admin).then(res => {
                        // Les rajoute un par un, sauf pour le supprimé
                        la.forEach(id => {
                            if (id!=uid) { this.addGroupAdmin(id, gid); }
                        });
                    });
                }
            }).then(res => { return true; }));
        });
    }
    
    //------------------------------------------------------------------------------------------------------------------------
    // Fonction d'édition TBT
    //------------------------------------------------------------------------------------------------------------------------
    
    /**
     * @summary Fonction qui édite un groupe existant dans le LDAP. Très similaire à {@link User.addGroup}
     * @desc Appelle {@link LDAP.add} bien sûr, mais aussi {@link Admin.addGroupMember} et {@link Admin.addGroupAdmin} en godmode pour gérer les groupes du nouvel utilisateur.
     * @arg {string} gid - Identifiant du groupe à modifier
     * @arg {Object} data - Dictionnaire des informations utilisateurs au même format que pour {@link User.addGroup} avec tous les champs optionnels.
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     */
    editGroup(gid, data) {
        // Récupération des anciennes données
        return this.peekGroup(gid).then(profil => {
            // Reecriture de profil avec les bons champs
            Object.keys(profil).forEach(keyLDAP => {
                Object.keys(config.group).forEach(keyAlias => {
                    config.group[keyAlias]=keyLDAP;
                    profil[keyAlias]=profil[keyLDAP];
                });
            });
            // Surcharge des champs à modifier selon data
            Object.keys(data).forEach(key => {
                profil[key]=data[key];
            });
            // Modification propre
            this.delGroup(gid).then(r => {
                this.addGroup(profil).then(res => { if (!res) { throw Error; } else { return true; }});
            });
        });
    }
    
    //------------------------------------------------------------------------------------------------------------------------
    // Fonctions de suppression TBT
    //------------------------------------------------------------------------------------------------------------------------
    
    /**
     * @summary Fonction qui supprime un groupe du LDAP.
     * @desc Cette fonction commence par gérer les groupes du membre puis le supprime entièrement.
     * Appelle {@link LDAP.clear} bien sûr, mais aussi {@link Admin.delGroupMember} et {@link Admin.delGroupAdmin} pour gérer les groupes de l'utilisateur sortant.
     * @arg {string} gid - gid de la victime
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     */
    delGroup(gid) {
        // Gestion des membres et administrateurs d'abord
        return this.constructor.peekGroup(gid).then(profil => {
            // Ordre important
            profil[config.group['admin']].forEach( id => { this.delGroupAdmin( id, gid); });
            profil[config.group['member']].forEach(id => { this.delGroupMember(id, gid); });
        // Elimination
        }).then(res => {
            LDAP.clear(config.key_id+"="+gid+","+config.dn_groups);
        });
    }
}

class SuperAdmin extends Admin {
    /**
     * @class Cette classe est la classe du super administrateur qui créé et supprime des membres.
     * @summary Bête appel au constructeur de la classe mère.
     * @arg {Object} user - Utilisateur de la forme nécessaire à {@link LDAP.bind}.
     * @author hawkspar
     */
    constructor(user) { super(user); }

    //------------------------------------------------------------------------------------------------------------------------
    // Fonctions de création TBT
    //------------------------------------------------------------------------------------------------------------------------
    
    /**
     * @summary Fonction qui créé un nouvel utilisateur dans le LDAP.
     * @desc Appelle {@link LDAP.add} bien sûr, mais aussi {@link User.addGroupMember} et {@link Admin.addGroupAdmin} pour gérer les groupes du nouvel utilisateur.
     * @arg {Object} data - Dictionnaire des informations utilisateurs. Des erreurs peuvent apparaître si tous les champs ne sont pas remplis.
     * @arg {string} data[givenName] - Prénom
     * @arg {string} data[lastName] - Nom
     * @arg {string} data[nickname] [] - Surnom
     * @arg {string} data[birthdate] - Date d'anniversaire au format annee-mois-jour
     * @arg {string} data[password] - Mot de passe
     * @arg {string} data[promotion] - Simple année d'entrée école (pas d'X17)
     * @arg {string} data[mail] - Courriel supposé valide
     * @arg {string} data[phone] - String du numéro de portable
     * @arg {string} data[photo] - Photo jpeg directement en bytestring
     * @arg {string} data[room] - Pour l'instant juste le numéro de casert ou rien, à terme l'adresse complète
     * @arg {string} data[readPerm] [] - Permissions spéciales BR sous la forme "*truc,"
     * @arg {string} data[writePerm] [] - Permissions spéciales BR sous la forme "*truc,"
     * @arg {string[]} data[forlifes] [] - Liste d'alias spéciaux BR (attention le filtre .fkz n'est plus fonctionnel)
     * @arg {string[]} data[ips] - Liste des ips connus de la personne
     * @arg {string[]} data[groups] [] - Liste des gid dont le pax est membre (doit contenir "on_platal" si il y a lieu)
     * @arg {string[]} data[groupsIsAdmin] [] - Liste des gid dont le pax est admin ; supposé sous-liste du précédent
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     */
    addUser(data) {
        // Calcul d'un dictionnaire d'ajout
        let vals = {};

        // uid de base généré à partir de nom et prénom, plus potentiellement promo et un offset
        // MEF mélange de Promise et de fonction standard
        Tests.generateUid(data['givenName'],data['lastName'],data['promotion']).then(id => { vals[config.key_id]=id; } );

        // Ecriture de toutes les valeurs directement inscrites dans le LDAP (in pour input)
        // Génère une erreur si un champ n'est pas rempli
        config.user.direct_input.forEach(key_att => vals[config.user[key_att]]=data[key_att]);

        // Appel à la fonction de base
        return LDAP.add(config.key_id+"="+vals[config.key_id]+","+config.dn_users, vals).then( res => {
            // Modifications multiples pour avoir plusieurs champs de même type ; boucle sur les attributs multiples (d'où mul)
            config.user.muliple_input.forEach(key_att => {
                // On rajoute chaque valeur en entrée
                data[key_att].forEach(val => {
                    let vals2 = {};
                    vals2[config.user[key_att]]=val;
                    LDAP.change(config.key_id+"="+vals[config.key_id]+","+config.dn_users, "add", vals2).then(res => {
                        if (!res) { throw Error; }
                    });
                });
            });

            // Certains champs nécessitent de petits calculs
            let vals3={};

            // Création d'un nom complet lisible
            vals3[config.user['fullName']]=data['givenName']+' '+data['lastName'].toUpperCase();
            // ?!
            vals3[config.user['password']] =  '{CRYPT}' + data['password'];
            // Ecriture d'un surnom s'il y a lieu
            if ((data['nickname']!=undefined)&(data['nickname']!='')) {
                vals3[config.user['nickname']]=data['nickname'];
            }
            // Génération id aléatoire unique
            Tests.generateId(config.user['id'], config.dn_users).then(id => { vals3[config.user['id']]=id; });
            
            // Stockage machine ; dépend du prénom
            vals3[config.user['directory']] = '/hosting/users/' + data['givenName'][0];

            // Code root
            vals3[config.user['cleanFullName']]=data['fullName'].replace(':', ';').toLowerCase().normalize('UFD');
            
            // Adressage root
            if (data['groups'].includes("on_platal")) { vals3[config.user['login']] = "/bin/bash"; }
            else  { vals3[config.user['login']] = "/sbin/nologin"; }
            
            // Permissions BR
            vals3[config.user['readPerm']] = 'br.*,public.*';
            if (data['readPerm'].length>0) { vals3[config.user['readPerm']] += ',' + data['readPerm']; }
            vals3[config.user['writePerm']] = 'br.*,!br.blague-du-jour,public.*,!br.campagnekes';
            if (data['writePerm'].length>0) { vals3[config.user['readPerm']] += ',' + data['writePerm']; }

            // Valeur nécessaire mais inutile
            vals3[config.user['idNum']] ='5000';

            // Inscription des valeurs calculées
            LDAP.change(config.key_id+"="+vals[config.user['hruid']]+","+config.dn_users, "add", vals3).then(res => {
                if (!res) { throw Error; }
            });

            ["posixAccount", "shadowAccount", "inetOrgPerson", "brAccount"].forEach(cst => {
                let val3={};
                vals3[config.user['class']]=cst;
                LDAP.change(config.key_id+"="+vals[config.user['hruid']]+","+config.dn_users, "add", vals3).then(res => {
                    if (!res) { throw Error; }
                });
            });

            // Utilisation des fonctions adaptées pour assurer la cohérence de l'ensemble
            data['groupsIsMember'].forEach(gid => { this.addGroupMember(vals[config.key_id], gid); });
            data['groupsIsAdmin'].forEach(gid => {  this.addGroupAdmin( vals[config.key_id], gid); });

            return true;
        });
    }

    //------------------------------------------------------------------------------------------------------------------------
    // Fonctions de suppression TBT
    //------------------------------------------------------------------------------------------------------------------------
    
    /**
     * @summary Fonction qui supprime un utilisateur du LDAP.
     * @desc Cette fonction commence par gérer les groupes du membre puis le supprime entièrement.
     * Appelle {@link LDAP.clear} bien sûr, mais aussi {@link Admin.delGroupMember} et {@link Admin.delGroupAdmin} pour gérer les groupes de l'utilisateur sortant.
     * @arg {string} uid - uid de la victime
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     */
    delUser(uid) {
        // Gestion des groupes d'abord
        return super.peekUser(uid).then(profil => {
            profil[config.user['groups']].forEach(gid => {
                this.isGroupAdmin(gid).then(res => {
                    this.delGroupAdmin(uid, gid);
                }).then(res => { this.delGroupMember(uid, gid); });
            });
        // Elimination
        }).then(res => { this.LDAP.clear(config.key_id+"="+uid+","+config.dn_users); });
    }
}

export {Admin, SuperAdmin};