/**
 * @file Ce fichier regroupe les différentes classes avec différents utilisateurs. Ces classes sont dédiées à être exportées directement pour être utilisées par le solver.
 * Le découpage par fichier est arbitraire mais permet de regrouper certaines classes proches.
 * @author hawkspar
 */

import LDAP from './basics';
import {SmartSearch, Tests} from './utilities';
import {Admin, SuperAdmin} from './admins';
// Essentiels pour le fichier de config
import path from 'path';
import fs from 'fs';

// Point central ; tous les champs de la BDD sont 'cachés' dans config.json et pas visibles directement
var configPath = path.resolve('./','ldap_config.json');
var config = JSON.parse(fs.readFileSync(configPath, 'utf8'));

//------------------------------------------------------------------------------------------------------------------------
// Classes à exporter TBT
//------------------------------------------------------------------------------------------------------------------------

class Open {
    /**
     * @class Cette classe est la classe exportable de base permettant à un utilisateur non connecté de faire des petites recherches simples.
     * @summary Constructeur vide.
     * @author hawkspar
    */
    constructor(config) {}

    //------------------------------------------------------------------------------------------------------------------------
    // Fonctions de lecture
    //------------------------------------------------------------------------------------------------------------------------
    /**
     * @summary Fonction qui retrouve les groupes dont un individu est membre.
     * @desc Cette fonction utilise {@link LDAP.search} va directement à la feuille de l'utilisateur.
     * @arg {string} uid - Identifiant de l'individu à interroger (le plus souvent prenom.nom, parfois l'année, supposé valide)
     * @return {Promise(string[])} Liste des uid de groupes (noms flat des groupes) où l'id fourni est membre
     * @static
     */
    static getGroups(uid) {
        return LDAP.search(config.key_id+uid+config.dn_users, config.user.groups)[0];
    }
    
    /**
     * @summary Fonction qui retrouve la liste des membres d'un groupe.
     * @desc Cette fonction utilise {@link LDAP.search} avec un dictionnaire prédéfini dans config.json.
     * @arg {string} gid - Identifiant du groupe à interroger (le plus souvent nom du groupe en minuscule)
     * @return {Promise(String[])} Liste des uid des membres où l'id fournie est membre (noms flat des groupes)
     * @static
     */
    static getMembers(gid) {
        return LDAP.search(config.key_id+gid+config.dn_users, config.group.member)[0];
    }
    
    /**
     * @summary Fonction qui retrouve la liste des admins d'un groupe.
     * @desc Cette fonction utilise {@link LDAP.search} avec un dictionnaire prédéfini dans config.json.
     * @arg {string} gid - Identifiant du groupe à interroger (le plus souvent nom du groupe en minuscule)
     * @return {Promise(string[])} Liste des uid des membres où l'id fournie est membre (noms flat des groupes)
     * @static
     */
    static getAdmins(gid) {
        return LDAP.search(config.key_id+gid+config.dn_users, config.group.admin)[0];
    }

    /**
     * @summary Cette fonction teste si un utilisateur est membre d'un groupe.
     * @desc Utilise les méthodes statiques {@link open.getGroups} et {@link open.getMembers}
     * @param {string} uid - Identifiant de l'utilisateur à tester 
     * @param {string} gid  - Identification du groupe à tester
     * @returns {Promise(boolean)} True si l'utilisateur est membre
     * @static
     */
    static isGroupMember(uid, gid) {
        return this.getGroups(uid).then(lg => {
            this.getMembers(gid).then(lm => {
                if (lg.includes(gid) && lm.includes(uid)) {
                    return true;
                }
                else {
                    return false;
                }
            });
        });
    }

    /**
     * @summary Cette fonction teste si un utilisateur est admin d'un groupe.
     * @desc Utilise la méthode statique {@link Open.getAdmins}
     * @param {string} uid - Identifiant de l'utilisateur à tester 
     * @param {string} gid  - Identification du groupe à tester
     * @returns {Promise(boolean)} True si l'utilisateur est administrateur
     * @static
     */
    static isGroupAdmin(uid, gid) {
        return this.getAdmins(gid).then(la => {
            if (la.includes(uid)) {
                return true;
            }
            else {
                return false;
            }
        });
    }
     
    /**
     * @summary Fonction qui renvoit toutes les infos relatives à un utilisateur particulier.
     * @desc Cette fonction utilise {@link LDAP.search} avec des attributs prédéfinis.
     * @arg {string} uid - Identifiant de l'utilisateur
     * @return {Promise(Object[])} Informations recueillies ; renvoie une liste de dictionnaire avec le profil complet de l'utilisateur ;
     * voir `ldap_config.json`(..\..\ldap_config.json) pour les clés exactes.
     * @static
     */
    static peekUser(uid) {
        return LDAP.search(config.key_id+uid+config.dn_users, config.user.profil);
    }
     
    /**
     * @summary Fonction qui renvoit toutes les infos relatives à un groupe particulier.
     * @desc Cette fonction utilise {@link LDAP.search} avec des attributs prédéfinis.
     * @arg {string} gid - Identifiant du groupe
     * @return {Promise(Object[])} Informations recueillies ; renvoie une liste de dictionnaire avec le profil complet du groupe ;
     * voir `ldap_config.json`(..\..\ldap_config.json) pour les clés exactes.
     * @static
     */
    static peekGroup(gid) {
        return LDAP.search(config.key_id+gid+config.dn_groups, config.group.profil);
    }

    //------------------------------------------------------------------------------------------------------------------------
    // Fonctions de recherche
    //------------------------------------------------------------------------------------------------------------------------

    /**
     * @summary Fonction qui retrouve le groupe qui ressemblent à l'input et qui correspond au type fourni. Etape 0 vers un vrai TOL (Trombino On Line).
     * @desc Cette fonction utilise {@link SmartSearch.groups}.
     * @arg {string} input - String entré par l'utilisateur qui ressemble au nom du groupe.
     * @arg {string} type - String aux valeurs prédéfinies dans ldap_config.
     * @return {Promise(string[])} Liste des gid dont le nom ressemble à l'input.
     * @static
    */
    static findGroups(input, type) {
        // Trucs intelligents faits dans ./utilities
        return SmartSearch.groups(input, [config.key_id, config.group.type]).then(gList => {
            let gidtyList = [];
            gList.forEach(g => {
                // Si le groupe est du bon type on rajoute son gid
                if (g[config.group.type]==type) { gidtyList.push(g[config.key_id]); }
            });
            return gidtyList;
        });
    }

    /**
     * @summary Fonction qui retrouve les uid des paxs validant les critères de recherche. Autre étape vers vrai TOL (Trombino On Line). Doit être préféré à repliquerTOL
     * car moins gourmande envers le LDAP (utiliser {@link peekUser} au cas par cas après pour obtenir les vraies infos).
     * @desc Cette fonction utilise {@link SmartSearch.users}.
     * @arg {Object} data - Dictionnaire contenant les données nécessaires à {@link SmartSearch.groups}
     * @return {Promise(string[])} gids des profils qui "match" les critères proposés.
     * @static
     */
    static findUsers(data) {
        return SmartSearch.users(data, config.key_id);
    }
}

class User extends Open {
    /**
     * @class Cette classe est la classe de l'utilisateur connecté qui peut déjà créer un groupe et changer son profil.
     * @summary Ce constructeur appelle simplement {@link LDAP.bind}.
     * @arg {Object} user - Utilisateur de la forme nécessaire à {@link LDAP.bind}.
     * @author hawkspar
    */
    constructor(user) {
        super();
        LDAP.bind(user);
    }

    //------------------------------------------------------------------------------------------------------------------------
    // Fonction de création TBT
    //------------------------------------------------------------------------------------------------------------------------

    /**
     * @summary Fonction qui créé un nouveau groupe dans le LDAP.
     * @desc Cette fonction fait une utilisation massive d'eval pour anonymiser son code ; c'est mal et cela suppose que beaucoup de soins ont été pris lors de
     * l'escape de ses paramètres. Appelle {@link LDAP.add} et {@link LDAP.change}, mais aussi {@link Admin.addMemberGroup} et {@link Admin.addAdminGroup}
     * pour gérer les groupes du nouvel utilisateur. Attention une manip FOIREUSE est cachée dedans.
     * @arg {Object} data - Dictionnaire des informations utilisateurs (voir détail des champs dans config.json)
     * @arg {string} data[name] - Nom du groupe
     * @arg {string} data[ns] - Statut du groupe ; 'binet' ou 'free', càd ouvert à tous
     * @arg {string[]} data[members] - Liste des membres du groupe
     * @arg {string[]} data[admins] - Liste des admins du groupe ; supposée être une sous-liste de la précédente
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     */
    addGroup(data) {
        // Calcul d'un dictionnaire d'ajout
        let vals = {};

        // uid de base généré à partir du nom standardisé
        Tests.generateReadableId(data['name']).then(id => { vals[config.group['name']]=id; });

        // Ecriture de toutes les valeurs directement inscrites dans le LDAP (in pour input)
        config.group.direct_input.forEach(key_att => vals[config.group[key_att]]=data[key_att]);

        // Appel à la fonction de base
        return LDAP.add(config.key_id+"="+vals[config.group['name']]+","+config.dn_groups, vals).then( res => {
            // Certains champs nécessitent de petits calculs
            let vals2={};

            // Sauvegarde du nom (pour le cas où gid != data['name'])
            vals2[config.group["name"]]=data['name'];

            // ?!
            vals2[config.user['password']] =  '';

            // Génération id aléatoire et test contre le LDAP
            Tests.generateId(config.groups["idNumber"], config.dn_groups).then(id => { vals2[config.group['idNumber']]=id; });
            // FOIREUX : Hypothèse sur la structure du reste des données mais évite un test.assurerUnicite à deux variables
            vals2[config.group['idNumber2']]=vals2[config.group['idNumber']];
            
            // Stockage machine ; dépend du prénom
            vals2[config.group['directory']] = '/hosting/groups/'+vals[config.key_id];

            // Code root
            vals2[config.group['cleanFullName']]=data['name'].replace(':', ';').toLowerCase().normalize('UFD');
            
            // Adressage root
            vals2[config.group['login']] = "/sbin/nologin";
            
            // Permissions BR
            vals2[config.group['readPerm']] = '!*';
            vals2[config.group['writePerm']] = '!*';

            // Inscription des valeurs calculées
            LDAP.change(config.key_id+"="+vals[config.key_id]+","+config.dn_groups, "add", vals2).then(res => {
                if (!res) { throw Error; }
            });

            ["posixAccount", "posixGroup", "brAccount"].forEach(cst => {
                let vals3={};
                vals3[config.group['class']]=cst;
                LDAP.change(config.key_id+"="+vals[config.key_id]+","+config.dn_groups, "add", vals3).then(res => {
                    if (!res) { throw Error; }
                });
            });

            // Utilisation des fonctions adaptées pour assurer la cohérence de l'ensemble
            data['members'].forEach(uid => { Admin.addGroupMember(uid, vals[config.key_att]); });
            data['admins'].forEach(uid => {  Admin.addGroupAdmin( uid, vals[config.key_att]); });

            return true;
        });
    }

    //------------------------------------------------------------------------------------------------------------------------
    // Fonctions d'édition TBT
    //------------------------------------------------------------------------------------------------------------------------
    
    /**
     * @summary Fonction qui édite un utilisateur existant dans le LDAP. Très similaire à {@link creerUtilisateur}
     * @desc Appelle simplement {@link creerUtilisateur} et {@link supprimerUtilisateur} en godmode, plus {@link renseignerSurUtilisateur} pour les champs non fournis.
     * Ce choix a pour conséquence que l'ordre du dictionnaire de correspondance dans ldap_config est important. A modifier car donne trop de pouvoir à l'utilisateur.
     * @arg {string} uid - Utilisateur à modifier (le plus souvent le même, mais root possible)
     * @arg {Object} data - Dictionnaire des informations utilisateurs au même format que pour {@link creerUtilisateur} avec tous les champs optionnels ;
     * MEF ces valeurs vont écraser les précédentes.
     * attention toutes les clés de cette entrée seront modifiées dans le LDAP
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     */
    editUser(uid, data) {
        // Récupération des anciennes données
        return this.peekUser(uid).then(profil => {
            // Reecriture de profil avec les bons champs
            Object.keys(profil).forEach(keyLDAP => {
                Object.keys(config.user).forEach(keyAlias => {
                    config.user[keyAlias]=keyLDAP;
                    profil[keyAlias]=profil[keyLDAP];
                });
            });
            // Régénération du champ manquant dans profil
            this.getGroups(uid).then(lg => {
                profil['groupsIsAdmin']=[];
                lg.forEach(gid =>{
                    this.isGroupAdmin(uid, gid).then(res =>{
                        if (res) { profil['groupsIsAdmin'].push(gid); }
                    }).then(res => {
                        // Surcharge des champs à modifier selon data
                        Object.keys(data).forEach(key => {
                            profil[key]=data[key];
                        });
                        // Passage en godmode
                        var god = SuperAdmin({"uid":"", "password":""});
                        // Modification propre
                        god.delUser(uid).then(r => {
                            god.createUser(profil).then(res => { if (!res) { throw Error; } else { return true; }});
                        });
                    });
                });
            });
        });
    }
    
    destr() { LDAP.unbind(); }
}

export {Open, User};